﻿using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.Text.RegularExpressions;

namespace Compiler
{
    public class Program
    {
        // 測試主程式。
        static void Main(string[] args)
        {
            C3Parser parser = new C3Parser(IO.fileToText(args[0]));
            parser.parse();
        }
    }

    public class Regexp
    {
        // 傳回 text 中符合正規表示式 pattern 的所有段落。
        public static List<String> matches(String text, String pattern, int groupId)
        {
            List<String> rzList = new List<String>();
            Match match = Regex.Match(text, pattern);
            for (int i = 0; match.Success; i++)
            {
                rzList.Add(match.Groups[groupId].Value);
                match = match.NextMatch();
            }
            return rzList;
        }

        // 類似 Linux 中的 grep 指令，印出比對成功的所有字串
        public static void grep(String fileName, String pattern, int groupId)
        {
            Console.WriteLine("=============grep(" + fileName + "," + pattern + "," + groupId+")=============");
            String text = IO.fileToText(fileName);
            List<String> m = Regexp.matches(text, pattern, groupId);
            foreach (String s in m)
                Console.WriteLine(s);
        }
    }

    public class IO
    {
        // 讀取文字檔，以字串形式傳回。
        public static String fileToText(String filePath)        
        {
            StreamReader file = new StreamReader(filePath);
            String text = file.ReadToEnd();
            file.Close();
            return text;
        }
    }

    public class RegexScanner
    {
        public static List<String> scan(String text)
        {
            return Regexp.matches(text, "\".*?\""+@"|(\d+)|([a-zA-Z]\w*)|([\+\-\*/<=>!]+)|([^\s])", 0); // 取得 數字串, 字元串, 或單一字元
        }
    }

    public class C3Parser
    {
        String text;
        List<String> tokens;
        List<String> types;
        int idx = 0;
        String keywords = "|if|for|while|";
        String OP1 = "++|--";
        String OP2 = "+|-|*|/";
        String BOP = "==|!=|<=|>=|<|>";
        String ITEM = "id|number|string";
		Stack<String> stack;
        int forCount = 0;
        int varCount = 0;
        StringBuilder toCodeText = new StringBuilder();

        public C3Parser(String pText)
        {
            text = pText;
            tokens = RegexScanner.scan(text);
            types = new List<String>();
            for (int i = 0; i < tokens.Count; i++)
            {
                String type = tokenToType(tokens[i]);
                types.Add(type);
                debug("token["+i+"] =\t" + tokens[i] + "\t" + type);
            }
        }

		static String spaces = "                                                                                                                 ";
		public void debug(int level, String msg)
		{
			Console.WriteLine(spaces.Substring(0, level)+msg);
		}

		public void debug(String msg)
        {
            debug(0, msg);
        }

        public void error(String msg)
        {
            debug("Error : token["+idx+"]="+tokens[idx]);
            debug("   "+msg);
            throw new Exception(msg);
        }

        public String tokenToType(String token)
        {
            if (keywords.IndexOf("|" + token + "|") >= 0)
                return token;
            else if (token[0] == '\"')
                return "string";
            else if (Char.IsDigit(token[0]))
                return "number";
            else if (Char.IsLetter(token[0]))
                return "id";
            else
                return token;
        }

        public bool isNext(String pTypes)
        {
            if (idx >= types.Count) 
                return false;
            if (("|"+pTypes+"|").IndexOf("|"+types[idx]+"|")>=0)
                return true;
            else
                return false;
        }

		public String next(String type)
		{
            if (isNext(type))
            {
                debug(stack.Count, "token[" + idx + "] =\t" + tokens[idx] + "\t" + types[idx]);
                return tokens[idx++];
            }
            else
            {
                error("next token is not in type:" + type);
                return null;
            }
		}


        /* Parser 
        PROG	= BaseList
        FOR		= for (STMT; BEXP; STMT) BLOCK
        BLOCK 	= '{' BaseList '}'
        BaseList= { BASE }
        BASE 	= FOR | STMT;
        STMT 	= id '=' EXP  | id '(' ArgList ')' | id OP1
        ArgList = EXP { ',' EXP }
        EXP 	= ITEM OP2 ITEM | ITEM
        BEXP 	= EXP BOP EXP
        */

        /* Code Generator ==> PCode
        = 
        ++
        --
        +
        -
        *
        /
        CMP
        JMP
        JNO
        JYES
        CALL
        */
        // 注意：STMT 以上的層次直接產生程式碼，STMT 以下的部分則傳回結果變數。
        // 使用 "+-FOR"+forCount, "+-IF"+ifCount, "+-WHILE"+whileCount 等方式產生標記，才能在一開始就知道後續的跳躍標記。

        public void parse()
        {
            debug("================= Parse Begin =================");
			stack = new Stack<String>();
			parseBaseList();
            debug("================= Generate Code ===============");
            debug(toCodeText.ToString());
        }

		public void push(String term) 
		{
			debug(stack.Count, "+"+term);
			stack.Push(term);
		}

		public void pop(String term)
		{
			String topTerm = stack.Pop();
			debug(stack.Count, "-"+topTerm);
			if (!topTerm.Equals(term))
				error("pop term="+topTerm+" should be "+term);
		}

        public void putOp(StringBuilder toCode, String op, String arg1, String arg2, String result)
        {
            toCode.Append("\t" + op + "\t" + arg1 + "\t" + arg2 + "\t" + result + "\n");
        }

        public void putOp(String op, String arg1, String arg2, String result)
        {
            putOp(toCodeText, op, arg1, arg2, result);
        }

        public void put(String code)
        {
            toCodeText.Append(code);
        }

        public String nextTemp()
        {
            String var = "?T" + varCount;
            varCount++;
            return var;
        }

        // FOR = for (STMT; BEXP; STMT) BLOCK
        public void parseFor()
        {
			push("FOR");
            next("for");
            next("(");
            String stmtCode1 = parseStmt();
            put(stmtCode1);
            next(";");
            int tempForCount = forCount++;
            String forBeginLabel = "+FOR"+tempForCount;
            String forEndLabel = "-FOR"+tempForCount;
            put(forBeginLabel + ":\n");
            String bexpCode = parseBexp();
            put(bexpCode);
            putOp("JNO", forEndLabel, "", "");
            next(";");
            String stmtCode2 = parseStmt();
            next(")");
            parseBlock();
            put(stmtCode2);
            putOp("JMP", forBeginLabel, "", "");
            put(forEndLabel + ":\n");
			pop("FOR");
		}

        // BLOCK = '{' BaseList '}'
        public void parseBlock()
        {
            push("BLOCK");
            next("{");
            parseBaseList();
            next("}");
			pop("BLOCK");
		}

        // BaseList= { BASE }
        public void parseBaseList()
        {
            push("BaseList");
            while (idx < tokens.Count && !isNext("}"))
                parseBase();
			pop("BaseList");
		}

        // BASE = FOR | STMT;
        public void parseBase()
        {
            push("BASE");
            if (isNext("for"))
                parseFor();
            else
            {
                String stmtCode = parseStmt();
                put(stmtCode);
                next(";");
            }
			pop("BASE");
		}

        // STMT = id '=' EXP  | id '(' ArgList ')' | id OP1
        public String parseStmt()
        {
            StringBuilder stmtCode = new StringBuilder();
            push("STMT");
            String id = next("id");
            if (isNext("="))                                // id '=' EXP             --> ASSIGN
            {
                next("=");
                String expVar = parseExp();
                putOp(stmtCode, "= ", expVar, "", id);
            }
            else if (isNext("("))                           // id '(' ArgList ')'     --> Function Call
            {
                putOp(stmtCode, "CALL ", id, "", "");
                next("(");
                if (!isNext(")"))
                {
                    List<String> argList = parseArgList();
                    foreach (String arg in argList)
                        putOp(stmtCode, "PARAM", arg, "", "");
                }
                next(")");
            }
            else                                            // id OP1
            {
                String op1 = next(OP1);
                putOp(stmtCode, op1, id, "", "");
            }
			pop("STMT");
            return stmtCode.ToString();
		}

        // ArgList = EXP { ',' EXP }
        public List<String> parseArgList() {
            List<String> rzList = new List<String>();
            String evar = parseExp();
            rzList.Add(evar);
            while (isNext(","))
            {
                next(",");
                evar = parseExp();
                rzList.Add(evar);
            }
            return rzList;
        }

        // EXP = ITEM OP2 ITEM | ITEM
        public String parseExp()
        {
            push("EXP");
            String rzVar = null;
            String item1 = next(ITEM);
            if (isNext(OP2))
            {
                String op = next(OP2);
                String item2 = next(ITEM);
                String temp = nextTemp();
                putOp(op, item1, item2, temp);
                rzVar = temp;
            } else
                rzVar = item1;
			pop("EXP");
            return rzVar;
		}

        // BEXP = EXP BOP EXP
        public String parseBexp()
        {
            StringBuilder rzCode = new StringBuilder();
            push("BEXP");
            String var1 = parseExp();
            String bop = next(BOP);
            String var2 = parseExp();
            putOp(rzCode, bop, var1, var2, "");
			pop("BEXP");
            return rzCode.ToString();
		}
    }
}