﻿using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.Text.RegularExpressions;

namespace Compiler
{
    public class Compiler
    {
        // 測試主程式。
        static void Main(string[] args)
        {
            C3Parser parser = new C3Parser(IO.fileToText(args[0]));
            Tree parseTree = parser.parse();
            Coder coder = new Coder(parseTree);
            IO.debug(coder.ToString());
        }
    }

/*
MODULE  = #module [\w.]* '{' [^\{]? '}'
INCLUDE = #include [\w.]*;
REWRITE = #rewrite [\w.]* '{' RULE_LIST '}'
RULE_LIST = RULE { ; RULE }
RULE    = '{' regexp => toexp }
*/
    public class Preprocessor
    {
    }

    public class Regexp
    {
        // 傳回 text 中符合正規表示式 pattern 的所有段落。
        public static List<String> matches(String text, String pattern, int groupId)
        {
            List<String> rzList = new List<String>();
            Match match = Regex.Match(text, pattern);
            for (int i = 0; match.Success; i++)
            {
                rzList.Add(match.Groups[groupId].Value);
                match = match.NextMatch();
            }
            return rzList;
        }

        // 類似 Linux 中的 grep 指令，印出比對成功的所有字串
        public static void grep(String fileName, String pattern, int groupId)
        {
            Console.WriteLine("=============grep(" + fileName + "," + pattern + "," + groupId+")=============");
            String text = IO.fileToText(fileName);
            List<String> m = Regexp.matches(text, pattern, groupId);
            foreach (String s in m)
                Console.WriteLine(s);
        }
    }

    public class IO
    {
        // 讀取文字檔，以字串形式傳回。
        public static String fileToText(String filePath)        
        {
            StreamReader file = new StreamReader(filePath);
            String text = file.ReadToEnd();
            file.Close();
            return text;
        }

        public static void debug(int level, String msg)
        {
            Console.WriteLine(STR.spaces(level) + msg);
        }

        public static void debug(String msg)
        {
            debug(0, msg);
        }
    }

    public class RegexScanner
    {
        public static List<String> scan(String text)
        {
            return Regexp.matches(text, "(\".*?\")" + @"|(\d+)|([a-zA-Z]\w*)|([\+\-\*/<=>!\.]+)|('\\?.?')|([^\s])", 0); // 取得 數字串, 變數名稱, 關鍵字, 字元 或 單一符號
        }
    }

    public class C3Parser
    {
        String text;
        List<String> tokens;
        List<String> types;
        int idx = 0;
        String KEYWORDS = "if|for|while|else|foreach|class|is|return|throw";
        //        String BASETYPE = "int|float|byte|char|string";|yield|rename|new
        String OP1 = "+|-|!";
        String OP_INC = "++|--";
        String OP2 = "+|-|==|!=|<=|>=|<|>";
        String OP2_MUL = "*|/";
		Stack<Tree> stack;

        public C3Parser(String pText)
        {
            text = pText;
            tokens = RegexScanner.scan(text);
            types = new List<String>();
            for (int i = 0; i < tokens.Count; i++)
            {
                String type = tokenToType(tokens[i]);
                types.Add(type);
                IO.debug("token["+i+"] =\t" + tokens[i] + "\t" + type);
            }
        }

        public void error(String msg)
        {
            IO.debug("Error : token["+idx+"]="+tokens[idx]);
            IO.debug("   "+msg);
            throw new Exception(msg);
        }

        public String tokenToType(String token)
        {
            if (("|" + KEYWORDS + "|").IndexOf("|" + token + "|") >= 0)
                return token;
            else if (token[0] == '\"')
                return "string";
            else if (Char.IsDigit(token[0]))
                return "number";
            else if (Char.IsLetter(token[0]))
                return "id";
            else
                return token;
        }

        public bool isNext(String pTypes)
        {
            if (idx >= tokens.Count) return false;
            return STR.isFoundAt(types[idx], pTypes);
        }

		public String next(String pTypes)
		{
            if (isNext(pTypes))
            {
//                if (("|"+KEYWORDS+"|.|{|}|(|)|[|]|;|,|").IndexOf(types[idx])<0)
                    addChild(types[idx], tokens[idx]);
                IO.debug(stack.Count, "token[" + idx + "] =\t" + tokens[idx] + "\t" + types[idx]);
                return tokens[idx++];
            }
            else
            {
                error("next token "+tokens[idx]+" is not in type:" + pTypes);
                return null;
            }
		}

		public Tree push(String pType) 
		{
            IO.debug(stack.Count, "+" + pType);
            Tree tree = new Tree(pType);
            stack.Push(tree);
            return tree;
		}

        public Tree pop(String pType)
		{
			Tree tree = stack.Pop();
            IO.debug(stack.Count, "-" + tree.type);
            if (!tree.type.Equals(pType))
				error("pop type="+tree.type+" should be "+pType);
            if (stack.Count > 0)
                stack.Peek().addChild(tree);
            return tree;
		}

        public Tree addChild(String pType, String pValue)
        {
            Tree child = new Tree(pType, pValue);
            stack.Peek().addChild(child);
            return child;
        }

        public String getForwardTo(String pTypes)
        {
            String rzStr = "";
            for (int i = idx; i < tokens.Count; i++)
            {
                rzStr += tokens[i];
                if (idx < tokens.Count - 1 && Char.IsLetter(types[idx][0]) && Char.IsLetter(types[idx + 1][0]))
                    rzStr += " ";
                if (STR.isFoundAt(types[i], pTypes))
                    break;
            }
            return rzStr;
        }

        // PROG	= BASE_LIST										FIRST(BASE)						\0
        public Tree parse()
        {
            IO.debug("================= Parse Begin =================");
            stack = new Stack<Tree>();
            Tree progTree = new Tree("PROG");
            stack.Push(progTree);
            parseBaseList();
            IO.debug("================= Parse Tree ===============");
            IO.debug(progTree.ToString());
            if (stack.Count == 1)
                return stack.Pop();
            else
            {
                error("parse fail : stack.count = " + stack.Count);
                return null;
            }
        }

        // BASE_LIST = { BASE }									FIRST(BASE)						}
        public void parseBaseList()
        {
            push("BaseList");
            while (idx < tokens.Count && !isNext("}"))
                parseBase();
            pop("BaseList");
        }

        // BASE   = IF | FOR | FOR_EACH | WHILE | TRY	 			FIRST(BASE)						if for foreach while try return yield rename { class id
        //          | BLOCK | CLASS | OBJ | METHOD | STMT
        // STMT   = INC; | ASSIGN ;
        // OBJECT = TYPE [id] BLOCK								id id							FOLLOW(BASE)
        // METHOD = TYPE id '(' EXP_LIST ')' BLOCK				id (							FOLLOW(BASE)
        // INC	  = MEMBER OP_INC 
        // ASSIGN = [CALL_LIST = ] EXP_LIST ;					id =							FOLLOW(BASE)
        public void parseBase()
        {
            push("BASE");
            if (isNext("if"))
                parseIf();
            else if (isNext("for"))
                parseFor();
            else if (isNext("foreach"))
                parseForeach();
            else if (isNext("while"))
                parseWhile();
            else if (isNext("try"))
                parseTry();
            else if (isNext("{"))
                parseBlock();
            else if (isNext("class"))
                parseClass();
            else if (isNext("return"))
                parseReturn();
            else // OBJECT or METHOD or INC or ASSIGN
            {
                String head = getForwardTo(";|{");
                int eqIdx = head.IndexOf("=");
                int pairIdx = head.IndexOf("(");
                if (eqIdx >= 0 && (pairIdx < 0 || eqIdx < pairIdx))  // = before ( , ex : a = f(x); it's a ASSIGN
                {
                    parseAssign();
                    next(";");
                }
                else if (head.IndexOf(" ") >= 0)
                {
                    if (head.IndexOf("(") >= 0)
                        parseMethod();
                    else
                        parseObject();
                }
                else
                {
                    parseExp();
                    next(";");
                }
            }
            pop("BASE");
        }

        // INC		= MEMBER OP_INC
        public void parseInc()
        {
            push("INC");
            parseMember();
            next(OP_INC);
            pop("INC");
        }

        // ASSIGN  = EXP_LIST | MEMBER_LIST =  EXP_LIST 		id =							FOLLOW(BASE)
        public void parseAssign()
        {
            push("ASSIGN");
            parseMemberList();
            next("=");
            parseExpList();
            pop("ASSIGN");
        }

        // OBJECT	= MEMBER [id] BLOCK								id id							FOLLOW(BASE)
        public void parseObject()
        {
            push("OBJECT");
            parseMember();
            if (isNext("id"))
                next("id");
            parseBlock();
            pop("OBJECT");
        }

        // METHOD  = MEMBER id '(' EXP_LIST ')' BLOCK				id (							FOLLOW(BASE)
        public void parseMethod()
        {
            push("METHOD");
            parseMember();
            next("id");
            next("(");
            parseExpList();
            next(")");
            parseBlock();
            pop("METHOD");
        }

        // IF      = if (EXP) BASE { else IF } [else BASE]			if								FOLLOW(BASE)
        public void parseIf()
        {
            push("IF");
            next("if");
            next("(");
            parseExp();
            next(")");
            parseBase();
            while (isNext("else"))
            {
                next("else");
                if (isNext("if"))
                    parseIf();
                else
                {
                    parseBase();
                    break;
                }
            }
            pop("IF");
        }

        // FOR		= for (STMT; EXP; STMT) BASE					for								FOLLOW(BASE)
        public void parseFor()
        {
            push("FOR");
            next("for");
            next("(");
            parseAssign();
            next(";");
            parseExp();
            next(";");
            parseInc();
            next(")");
            parseBase();
            pop("FOR");
        }

        // FOR_EACH= foreach (id in EXP) BASE						foreach							FOLLOW(BASE)
        public void parseForeach()
        {
            push("FOR_EACH");
            next("foreach");
            next("(");
            next("id");
            next("in");
            parseExp();
            next(")");
            parseBase();
            pop("FOR_EACH");
        }

        // WHILE   = while (EXP) BASE								while							FOLLOW(BASE)
        public void parseWhile()
        {
            push("WHILE");
            next("while");
            next("(");
            parseExp();
            next(")");
            parseBase();
            pop("WHILE");
        }

        // TRY		= try BLOCK { catch '(' MEMBER ')' BLOCK }			try								FOLLOW(BASE)
        public void parseTry()
        {
            push("TRY");
            next("try");
            parseBlock();
            if (isNext("catch"))
            {
                next("catch");
                next("(");
                parseMember();
                next(")");
                parseBlock();
            }
            pop("TRY");
        }

        // BLOCK 	= '{' BASE_LIST '}'								{								FOLLOW(BASE)
        public void parseBlock()
        {
            push("BLOCK");
            next("{");
            parseBaseList();
            next("}");
            pop("BLOCK");
        }

        // CLASS   = class id is MEMBER BLOCK								class							FOLLOW(BASE)
        public void parseClass()
        {
            push("CLASS");
            next("class");
            next("id");
            if (isNext("is"))
            {
                next("is");
                parseMember();
            }
            parseBlock();
            pop("CLASS");
        }

        // MEMBER_LIST = [MEMBER] { , MEMBER }						id								= 
        public void parseMemberList()
        {
            push("MEMBER_LIST");
            if (!isNext("="))
                parseMember();
            while (isNext(","))
            {
                next(",");
                parseMember();
            }
            pop("MEMBER_LIST");
        }

        // EXP_LIST= [EXP] { , EXP }									FIRST(EXP)						) ;
        public void parseExpList()
        {
            push("EXP_LIST");
            if (!(isNext(";") || isNext(")")))              // FOLLOW = ; )
                parseExp();
            while (isNext(","))
            {
                next(",");
                parseExp();
            }
            pop("EXP_LIST");
        }

        // EXP	= OP1 EXP | MEXP { OP2 MEXP}				number string id PRE_OP1 (		POST_OP1 ) ;
        public void parseExp()
        {
            push("EXP");
            if (isNext(OP1))
            {
                next(OP1);
                parseExp();
            }
            else
            {
                parseMexp();
                while (isNext(OP2))
                {
                    next(OP2);
                    parseMexp();
                }
            }
            pop("EXP");
        }

        // MEXP    = ( EXP ) |  ITEM { OP2_MUL ITEM }				number string id (				OP2 ) ;
        public void parseMexp()
        {
            push("MEXP");
            if (isNext("("))
            {
                next("(");
                parseExp();
                next(")");
            }
            else
            {
                parseItem();
                while (isNext(OP2_MUL))
                {
                    next(OP2_MUL);
                    parseItem();
                }
            }
            pop("MEXP");
        }

        // RANGE   = ITEM .. ITEM
        public void parseRange() 
        {
            push("RANGE");
            parseItem();
            next("..");
            parseItem();
            pop("RANGE");
        }

        // LIST    = '[' { ITEM } ']'
        public void parseList()
        {
            push("LIST");
            if (isNext("["))
            {
                while (!isNext("]"))  {
                    next("]");
                    parseItem();
                }
            }
            pop("LIST");
        }

        // ITEM	= MEMBER | CALLS | LIST | number | char | string				number string id				OP2_MUL ) ;
        public void parseItem()
        {
            push("ITEM");
            if (isNext("number"))
                next("number");
            else if (isNext("string"))
                next("string");
            else if (isNext("char"))
                next("char");
            else if (isNext("["))
                parseList();
            else
                parseCalls();
            pop("ITEM");
        }

        // CALLS    = ONE_CALL { . ONE_CALL }								id								POST_OP1 OP2 ) ;
        public void parseCalls()
        {
            push("CALLS");
            parseOneCall();
            while (isNext(".")) 
            {
                next(".");
                parseOneCall();
            }
            pop("CALLS");
        }

        // ONE_CALL   = id ['(' EXP_LIST ')']							id								. POST_OP1 OP2 ) ;
        public void parseOneCall()
        {
            push("ONE_CALL");
            next("id");
            if (isNext("("))
            {
                next("(");
                parseExpList();
                next(")");
            }
            pop("ONE_CALL");
        }

        // MEMBER  = id { . id }									id								id
        public void parseMember()
        {
            push("MEMBER");
            next("id");
            while (isNext("."))
            {
                next(".");
                next("id");
            }
            pop("MEMBER");
        }

        // RETURN  = return EXP_LIST;								return							}
        public void parseReturn()
        {
            push("RETURN");
            next("return");
            parseExpList();
            next(";");
            pop("RETURN");
        }
    }

    public class Tree // Abstract Syntax Tree
    {
        public String type = null;
        public String value = null;
        public List<Tree> childs = null;

        public Tree(String pType) { type = pType;  }
        public Tree(String pType, String pValue) { type = pType; value = pValue; }

        public void addChild(Tree child)
        {
            if (childs == null)
                childs = new List<Tree>();
            childs.Add(child);
        }

        public static String toText(Tree tree, int level)
        {
            StringBuilder rzText = new StringBuilder();
            rzText.Append(STR.spaces(level));
            rzText.Append("+"+tree.type);
            if (tree.value != null)
                rzText.Append("\t = "+tree.value);
            rzText.Append("\n");
            if (tree.childs != null)
            {
                foreach (Tree child in tree.childs)
                    rzText.Append(toText(child, level + 1));
                rzText.Append(STR.spaces(level) + "-" + tree.type + "\n");
            }
            return rzText.ToString();
        }

        public override String ToString()
        {
            return toText(this, 0);
        }
    }

    public class STR
    {
        static String spacesText = "                                                                                                                 ";
        public static String spaces(int n) { return spacesText.Substring(0, n); }
//        public static String spaces(int n) { return String.Format("{0:"+n+"}", ""); }
        public static bool isFoundAt(String pItem, String pItemList)
        {
            if (("|" + pItemList + "|").IndexOf("|" + pItem + "|") >= 0)
                return true;
            else
                return false;
        }
    }

    public class Coder
    {
        public StringBuilder toCodeText = new StringBuilder();
        public Tree tree;
        public Stack<StringBuilder> stack = new Stack<StringBuilder>();

        public Coder(Tree pTree) 
        {
            tree = pTree;
            push();
            toCode(tree);
            pop();
        }

        public void output(String str)
        {
            toCodeText.Append(str);
            Console.Write(str);
        }

        public void push() { stack.Push(new StringBuilder()); }
        public String pop() { return stack.Pop().ToString(); }
        public void append(String s) { stack.Peek().Append(s + " "); }

        public void toCode(Tree tree)
        {
            if (tree.value != null)
            {
                if (tree.value.Equals("{"))
                {
                    output(STR.spaces(stack.Count - 1) + pop() + "\n");
                    push();
                    output(STR.spaces(stack.Count - 1) + "{\n");
                    push();
                }
                else if (tree.value.Equals("}"))
                {
                    pop();
                    output(STR.spaces(stack.Count - 1) + "}\n");
                }
                else if (tree.value.Equals(";"))
                {
                    String line = pop();
                    push();
                    output(STR.spaces(stack.Count - 1) + line + ";\n");
                }
                else
                    append(tree.value);
            }
            if (tree.childs != null)
            {
                foreach (Tree child in tree.childs)
                    toCode(child);
            }
        }

        public override String ToString() 
        { 
            return toCodeText.ToString(); 
        }
    }

    /* Code Generator ==> PCode
    = 
    ++
    --
    +
    -
    *
    /
    CMP
    JMP
    JNO
    JYES
    CALL
    */
    // 注意：STMT 以上的層次直接產生程式碼，STMT 以下的部分則傳回結果變數。
    // 使用 "+-FOR"+forCount, "+-IF"+ifCount, "+-WHILE"+whileCount 等方式產生標記，才能在一開始就知道後續的跳躍標記。
/*
    public abstract class Coder
    {
        public int forCount = 0;
        public int varCount = 0;
        public StringBuilder toCodeText = new StringBuilder();

        public Coder() {}

        public abstract String toCode(Tree tree);

        public String nextTemp()
        {
            String var = "?T" + varCount;
            varCount++;
            return var;
        }

        public static String pcode(String op, String arg1, String arg2, String result)
        {
            return "\t" + op + "\t" + arg1 + "\t" + arg2 + "\t" + result + "\n";
        }

        public static void putOp(StringBuilder toCode, String op, String arg1, String arg2, String result)
        {
            toCode.Append("\t" + op + "\t" + arg1 + "\t" + arg2 + "\t" + result + "\n");
        }

        public void putOp(String op, String arg1, String arg2, String result)
        {
            putOp(toCodeText, op, arg1, arg2, result);
        }

        public void put(String code)
        {
            toCodeText.Append(code);
        }

        public override String ToString() { return toCodeText.ToString(); }
    }

    public class PCoder : Coder
    {
        public void childsToCode(Tree tree)
        {
            foreach (Tree child in tree.childs)
                toCode(child);
        }

        public override String toCode(Tree tree) 
        {
            if (tree == null) return null;

            String type = tree.type;
            List<Tree> childs = tree.childs;

            if (type.Equals("PROG"))           // PROG	= BaseList
                childsToCode(tree);
            else if (type.Equals("FOR"))       // FOR	= for (STMT; BEXP; STMT) BASE
            {
                String stmtCode1 = toCode(childs[0]); // STMT
                put(stmtCode1);
                int tempForCount = forCount++;
                String forBeginLabel = "+FOR"+tempForCount;
                String forEndLabel = "-FOR"+tempForCount;
                put(forBeginLabel + ":\n");
                String bexpCode = toCode(childs[1]); // BEXP
                put(bexpCode);
                putOp("JF", forEndLabel, "", "");
                String stmtCode2 = toCode(childs[2]); // STMT
                put(stmtCode2);
                toCode(childs[3]); // BASE
                putOp("JMP", forBeginLabel, "", "");
                put(forEndLabel + ":\n");
            }
            else if (type.Equals("BLOCK"))     // BLOCK 	= '{' BaseList '}'
                childsToCode(tree);
            else if (type.Equals("BaseList"))  // BaseList= BASE*
                childsToCode(tree);
            else if (type.Equals("BASE"))      // BASE 	= FOR | STMT;
                childsToCode(tree);
            else if (type.Equals("STMT"))      // STMT 	= id '=' EXP  | id '(' ArgList ')' | id OP1
            else if (type.Equals("ASSIGN"))      // STMT 	= id '=' EXP  | id '(' ArgList ')' | id OP1
            else if (type.Equals("CALL"))      // STMT 	= id '=' EXP  | id '(' ArgList ')' | id OP1
            else if (type.Equals("OP1STMT"))      // STMT 	= id '=' EXP  | id '(' ArgList ')' | id OP1
            // STMT = ASSIGN  | CALL | OP1STMT
            // ASSIGN = id '=' EXP
            // CALL = id '(' ArgList ')'
            // OP1STMT = id OP1
            else if (type.Equals("STMT"))      // STMT 	= id '=' EXP  | id '(' ArgList ')' | id OP1
            {
                StringBuilder stmtCode = new StringBuilder();
                String id = childs[0].value;
                if (childs[1].type.Equals("=")) // id '=' EXP
                {
                    String expVar = toCode(childs[2]);
                    putOp(stmtCode, "=", expVar, "", id);
                }
                else if (childs[1].type.Equals("(")) // id '(' ArgList ')' 
                {
                    List<String> argList = toArgList(childs[2]);
                    putOp(stmtCode, "CALL", id, "", "");
                    foreach (String arg in argList)
                        putOp(stmtCode, "PARAM", arg, "", "");
                }
                else // id OP1
                {
                    String op1 = childs[1].value;
                    putOp(stmtCode, op1, id, "", "");
                }
                return stmtCode.ToString();
            }
//            else if (type.Equals("ArgList"))   // ArgList = EXP (',' EXP)*
            else if (type.Equals("EXP"))        // EXP 	= ITEM OP2 ITEM | ITEM
            {
                String rzVar;
                if (childs.Count == 1)          // EXP = ITEM
                {
                    rzVar = toCode(childs[0]);
                }
                else                            // EXP = ITEM OP2 ITEM
                {
                    String item1 = toCode(childs[0]);       // ITEM
                    String op = toCode(childs[1]);          // OP2
                    String item2 = toCode(childs[2]);       // ITEM
                    String temp = nextTemp();
                    putOp(op, item1, item2, temp);
                    rzVar = temp;
                }
                return rzVar;
            }
            else if (type.Equals("BEXP"))      // BEXP 	= EXP BOP EXP
            {
                String var1 = toCode(childs[0]);
                String bop  = childs[1].value;
                String var2 = toCode(childs[2]);
                return pcode(bop, var1, var2, "");
            }
            return null;
        }

        public List<String> toArgList(Tree tree)
        {
            List<String> argList = new List<String>();
            foreach (Tree child in tree.childs)
                argList.Add(toCode(child));
            return argList;
        }
    }
                Tree node = push(nodeType);
                parseCalls(); 
                if (isNext(OP_INC))
                {
                    next(OP_INC);
                    next(";");
                    nodeType = "INC";
                }
                else if (isNext(",") || isNext("=") || isNext(";"))
                {
                    // ASSIGN  = [MEMBER_LIST = ] EXP_LIST ;						id =							FOLLOW(BASE)
                    while (isNext(","))
                    {
                        next(",");
                        parseMember();
                    }
                    if (isNext("="))
                        next("=");
                    while (!isNext(";"))
                        parseExpList();
                    next(";");
                    nodeType = "ASSIGN";
                    IO.debug("nodeType = "+nodeType);
                }
                else if (isNext("id"))
                {
                    //                  checkType(callList);
                    String type = null;
                    if (isNext("id"))
                        type = next("id");
                    if (isNext("{")) // OBJECT		= MEMBER [id] BLOCK								id id							FOLLOW(BASE)
                    {
                        parseBlock();
                        nodeType = "OBJECT";
                    }
                    else if (type.Equals("id") && isNext("("))  // METHOD  = MEMBER id '(' EXP_LIST ')' BLOCK				id (							FOLLOW(BASE)
                    {
                        next("(");
                        parseExpList();
                        next(")");
                        parseBlock();
                        nodeType = "METHOD";
                    }
                    else error("Error : in parseBase !");
                }
                node.type = nodeType;
                pop(nodeType);
*/
/*
                Tree node = push(nodeType);
                parseCalls(); 
                if (isNext(OP_INC))
                {
                    next(OP_INC);
                    next(";");
                    nodeType = "INC";
                }
                else if (isNext(",") || isNext("=") || isNext(";"))
                {
                    // ASSIGN  = [MEMBER_LIST = ] EXP_LIST ;						id =							FOLLOW(BASE)
                    while (isNext(","))
                    {
                        next(",");
                        parseMember();
                    }
                    if (isNext("="))
                        next("=");
                    while (!isNext(";"))
                        parseExpList();
                    next(";");
                    nodeType = "ASSIGN";
                    IO.debug("nodeType = "+nodeType);
                }
                else if (isNext("id"))
                {
                    //                  checkType(callList);
                    String type = null;
                    if (isNext("id"))
                        type = next("id");
                    if (isNext("{")) // OBJECT		= MEMBER [id] BLOCK								id id							FOLLOW(BASE)
                    {
                        parseBlock();
                        nodeType = "OBJECT";
                    }
                    else if (type.Equals("id") && isNext("("))  // METHOD  = MEMBER id '(' EXP_LIST ')' BLOCK				id (							FOLLOW(BASE)
                    {
                        next("(");
                        parseExpList();
                        next(")");
                        parseBlock();
                        nodeType = "METHOD";
                    }
                    else error("Error : in parseBase !");
                }
                node.type = nodeType;
                pop(nodeType);
*/
}

